{
  Copyright 2023-2023 Michalis Kamburelis.

  This file is part of "Castle Game Engine".

  "Castle Game Engine" is free software; see the file COPYING.txt,
  included in this distribution, for details about the copyright.

  "Castle Game Engine" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  ----------------------------------------------------------------------------
}

{ Main "playing game" view, where most of the game logic takes place. }
unit GameViewPlay;

interface

uses Classes,
  CastleComponentSerialize, CastleUIControls, CastleControls, CastleNotifications,
  CastleKeysMouse, CastleViewport, CastleScene, CastleVectors, CastleCameras,
  CastleTransform, GameEnemy;

type
  { Direction expressed a 4 possible horizontal directions.
    In the clockwise order. }
  TDirection = (dirNorth, dirEast, dirSouth, dirWest);

const
  { Convert TDirection to MainViewport.Camera.Direction vector.
    Check these values looking at MainViewport.Camera.Direction (in "All" tab)
    in editor. }
  DirectionVector: array [TDirection] of TVector3 = (
    (X:  0; Y: 0; Z: -1),
    (X:  1; Y: 0; Z:  0),
    (X:  0; Y: 0; Z:  1),
    (X: -1; Y: 0; Z:  0)
  );

  DirectionName: array [TDirection] of String = (
    'North',
    'East',
    'South',
    'West'
  );

  { Time, in seconds, to perform move or rotation. }
  ActionDuration = 0.25;

type
  { Main "playing game" view, where most of the game logic takes place. }
  TViewPlay = class(TCastleView)
  published
    { Components designed using CGE editor.
      These fields will be automatically initialized at Start. }
    LabelFps: TCastleLabel;
    MainViewport: TCastleViewport;
    MainFog: TCastleFog;
    ButtonMoveLeft: TCastleButton;
    ButtonMoveBackward: TCastleButton;
    ButtonMoveRight: TCastleButton;
    ButtonMoveForward: TCastleButton;
    ButtonRotateLeft: TCastleButton;
    ButtonRotateRight: TCastleButton;
    LabelDirection: TCastleLabel;
    Health: TCastleLabel;
    Ammo: TCastleLabel;
    MapCamera: TCastleCamera;
  private
    //MainNotifications: TCastleNotifications;
    Viewport1: TCastleViewport;


    { Always synchronized with MainViewport.Camera.Direction. }
    Direction: TDirection;

    { Calculate "Value + Increase"
      doing typecasting as necessary between enums and integers,
      and making sure result is in TDir range. }
    function IncreaseDirection(const Value: TDirection; const Increase: Integer): TDirection;

    procedure Move(const MoveDirection: TDirection);
    procedure Rotate(const RotationChange: Integer);
    procedure ClickMoveLeft(Sender: TObject);
    procedure ClickMoveBackward(Sender: TObject);
    procedure ClickMoveRight(Sender: TObject);
    procedure ClickMoveForward(Sender: TObject);
    procedure ClickRotateLeft(Sender: TObject);
    procedure ClickRotateRight(Sender: TObject);
  public

    Enemies: TEnemyList;
    constructor Create(AOwner: TComponent); override;
    procedure Start; override;
    procedure Stop; override;
    procedure Update(const SecondsPassed: Single; var HandleInput: Boolean); override;
    function Press(const Event: TInputPressRelease): Boolean; override;
  end;

var
  ViewPlay: TViewPlay;
  Health_P: Integer;
  Ammo_P: Integer;

implementation

uses SysUtils, Math, CastleSoundEngine,
  CastleLog, CastleStringUtils, CastleFilesUtils, CastleUtils,
  GameViewMenu;

{ TViewPlay ----------------------------------------------------------------- }

constructor TViewPlay.Create(AOwner: TComponent);
begin
  inherited;
  DesignUrl := 'castle-data:/gameviewplay.castle-user-interface';
end;

procedure TViewPlay.Start;
var
  SoldierScene: TCastleScene;
  Enemy: TEnemy;
  I: Integer;
begin
  inherited;

  Health_P := 100;
  Ammo_P := 50;
  Enemies := TEnemyList.Create(true);
   for I := 1 to 1 do
   begin
     SoldierScene := DesignedComponent('Enemy' + IntToStr(I)) as TCastleScene;
     { Below using nil as Owner of TEnemy, as the Enemies list already "owns"
       instances of this class, i.e. it will free them. }
     Enemy := TEnemy.Create(nil);
     SoldierScene.AddBehavior(Enemy);
     Enemies.Add(Enemy);
   end;

  //LabelFps := DesignHud.DesignedComponent('LabelFps') as TCastleLabel;
  //MainNotifications := DesignHud.DesignedComponent('MainNotifications') as TCastleNotifications;
  Viewport1 := DesignedComponent('Viewport1') as TCastleViewport;

  { Make sure Dir, MainViewport.Camera.Direction and LabelDirection are initially synchronized. }
  Direction := dirNorth;
  MainViewport.Camera.Direction := DirectionVector[Direction];
  LabelDirection.Caption := 'Looking at: ' + DirectionName[Direction];

  Viewport1.Items := MainViewport.Items;
  Viewport1.Camera := MapCamera;

  { At design-time, we keep MainFog.VisibilityRange larger,
    otherwise it makes it hard to actually see level at design-time from top. }
  //MainFog.VisibilityRange := 5;

  { Assign events }
  ButtonMoveLeft.OnClick := {$ifdef FPC}@{$endif} ClickMoveLeft;
  ButtonMoveBackward.OnClick := {$ifdef FPC}@{$endif} ClickMoveBackward;
  ButtonMoveRight.OnClick := {$ifdef FPC}@{$endif} ClickMoveRight;
  ButtonMoveForward.OnClick := {$ifdef FPC}@{$endif} ClickMoveForward;
  ButtonRotateLeft.OnClick := {$ifdef FPC}@{$endif} ClickRotateLeft;
  ButtonRotateRight.OnClick := {$ifdef FPC}@{$endif} ClickRotateRight;
end;

procedure TViewPlay.Stop;
begin
  FreeAndNil(Enemies);
  inherited;
end;

procedure TViewPlay.Update(const SecondsPassed: Single; var HandleInput: Boolean);
var
  DirectionHorizontal: TVector3;
  GameActive: Boolean;
  Health_L: TCastleLabel;
  Ammo_L: TCastleLabel;
begin
  inherited;

  Health_L := DesignedComponent('Health') as TCastleLabel;
  Ammo_L := DesignedComponent('Ammo') as TCastleLabel;

  Health_L.Caption := IntToStr(Health_P);
  Ammo_L.Caption := IntToStr(Ammo_P);
  //LabelFps.Caption := Container.Fps.ToString;
  { This virtual method is executed every frame (many times per second). }
  //Assert(LabelFps <> nil, 'If you remove LabelFps from the design, remember to remove also the assignment "LabelFps.Caption := ..." from code');
  //LabelFps.Caption := 'FPS: ' + Container.Fps.ToString;

  //GameActive := Container.FrontView = Self;

  //MainViewport.Items.Paused := not GameActive;

  //if GameActive then
  //begin
    DirectionHorizontal := MainViewport.Camera.Direction;
    //if not VectorsParallel(DirectionHorizontal, MainViewport.Camera.GravityUp) then
    //  MakeVectorsOrthoOnTheirPlane(DirectionHorizontal, MainViewport.Camera.GravityUp);

    Viewport1.Camera.SetView(
      MainViewport.Camera.WorldTranslation + Vector3(0, 10, 0),
      Vector3(0, -1, 0),
      DirectionHorizontal);
  //end;


  { Disable buttons to move/rotate when we're during camera animation.
    The Move/Rotate methods would ignore the call anyway,
    but disabling the buttons allows to show it visually. }
  ButtonMoveLeft.Enabled := not MainViewport.Camera.Animation;
  ButtonMoveBackward.Enabled := not MainViewport.Camera.Animation;
  ButtonMoveRight.Enabled := not MainViewport.Camera.Animation;
  ButtonMoveForward.Enabled := not MainViewport.Camera.Animation;
  ButtonRotateLeft.Enabled := not MainViewport.Camera.Animation;
  ButtonRotateRight.Enabled := not MainViewport.Camera.Animation;
end;

function TViewPlay.IncreaseDirection(const Value: TDirection; const Increase: Integer): TDirection;
begin
  Result := TDirection(ChangeIntCycle(Ord(Value), Increase, Ord(High(TDirection))));
end;

procedure TViewPlay.Move(const MoveDirection: TDirection);
const
  GridSize = 1;
var
  NewPos, Pos, Dir, Up: TVector3;
begin
  if MainViewport.Camera.Animation then
    Exit;
  MainViewport.Camera.GetWorldView(Pos, Dir, Up);
  NewPos := Pos + DirectionVector[MoveDirection] * GridSize;
  if not MainViewport.Items.WorldSegmentCollision(Pos, NewPos) then
    MainViewport.Camera.AnimateTo(NewPos, Dir, Up, ActionDuration);
end;

procedure TViewPlay.Rotate(const RotationChange: Integer);
var
  Pos, Dir, Up: TVector3;
begin
  if MainViewport.Camera.Animation then
    Exit;
  MainViewport.Camera.GetWorldView(Pos, Dir, Up);
  Direction := IncreaseDirection(Direction, RotationChange);
  LabelDirection.Caption := 'Looking at: ' + DirectionName[Direction];
  Dir := DirectionVector[Direction];
  MainViewport.Camera.AnimateTo(Pos, Dir, Up, ActionDuration);
end;

function TViewPlay.Press(const Event: TInputPressRelease): Boolean;
var
  HitEnemy: TEnemy;
  Pistol: TCastleScene;
begin



  Result := inherited;
  if Result then Exit; // allow the ancestor to handle keys

  if Event.IsMouseButton(buttonLeft) then
  begin
    Pistol := DesignedComponent('Pistol') as TCastleScene;
    Pistol.PlayAnimation('shoot', false);
    SoundEngine.Play(SoundEngine.SoundFromName('shoot_sound'));
    Dec(Ammo_P);

    { We clicked on enemy if
      - TransformUnderMouse indicates we hit something
      - It has a behavior of TEnemy. }
    if (MainViewport.TransformUnderMouse <> nil) and
       (MainViewport.TransformUnderMouse.FindBehavior(TEnemy) <> nil) then
    begin
      HitEnemy := MainViewport.TransformUnderMouse.FindBehavior(TEnemy) as TEnemy;
      HitEnemy.Hurt;
    end;

    Exit(true);
  end;

  if not MainViewport.Camera.Animation then
  begin
    if Event.IsKey(keyW) then
    begin
      Move(IncreaseDirection(Direction, 0)); // Move along Direction
      Exit(true);
    end;
    if Event.IsKey(keyS) then
    begin
      Move(IncreaseDirection(Direction, 2)); // Move along the opposite of Direction
      Exit(true);
    end;
    if Event.IsKey(keyD) then
    begin
      Move(IncreaseDirection(Direction, 1)); // Move along Direction rotated to right (i.e. strafe right)
      Exit(true);
    end;
    if Event.IsKey(keyA) then
    begin
      Move(IncreaseDirection(Direction, -1)); // Move along Direction rotated to left (i.e. strafe left)
      Exit(true);
    end;

    if Event.IsKey(keyQ) then
    begin
      Rotate(-1);
      Exit(true);
    end;
    if Event.IsKey(keyE) then
    begin
      Rotate(1);
      Exit(true);
    end;
  end;

  if Event.IsKey(keyF5) then
  begin
    Container.SaveScreenToDefaultFile;
    Exit(true);
  end;

  if Event.IsKey(keyEscape) then
  begin
    Container.View := ViewMenu;
    Exit(true);
  end;
end;

procedure TViewPlay.ClickMoveLeft(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, -1));
end;

procedure TViewPlay.ClickMoveBackward(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 2));
end;

procedure TViewPlay.ClickMoveRight(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 1));
end;

procedure TViewPlay.ClickMoveForward(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 0));
end;

procedure TViewPlay.ClickRotateLeft(Sender: TObject);
begin
  Rotate(-1);
end;

procedure TViewPlay.ClickRotateRight(Sender: TObject);
begin
  Rotate(1);
end;

end.
